/*******************************************************************************
 * Copyright 2008, 2012 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.connection.impl;

import java.io.BufferedInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.ConnectException;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.logging.Level;

import magma.util.connection.ConnectionException;

/**
 * Base for socket connection implementations
 * 
 * @author Klaus Dorer
 */
public class TCPConnection extends SocketConnection
{
	// Network socket
	private Socket socket;

	// Outgoing data stream
	protected DataOutputStream out;

	// Incoming data stream
	protected InputStream in;

	/** the time when receive started (in ns) */
	protected long startTime;

	protected long endTime;

	/**
	 * Default constructor
	 */
	public TCPConnection()
	{
		this(SERVER_IP, SERVER_PORT);
	}

	/**
	 * Assignment constructor
	 * 
	 * @param host Host address
	 * @param port Port number
	 */
	public TCPConnection(String host, int port)
	{
		super(host, port);
	}

	/**
	 * Actually connect to the server
	 */
	@Override
	public void establishConnection() throws ConnectionException
	{
		try {
			socket = new Socket(host, port);
			socket.setTcpNoDelay(true);

			in = new BufferedInputStream(socket.getInputStream());
			out = new DataOutputStream(socket.getOutputStream());

			// System.out.println("Connection to: " + host + ":" + port);

			connected = true;

		} catch (UnknownHostException e) {
			throw new ConnectionException(e.getMessage(), e);

		} catch (ConnectException e) {
			String message = host + ":" + port
					+ " refused the connection. Is server running?";
			throw new ConnectionException(message, e);

		} catch (IOException e) {
			e.printStackTrace();
			throw new ConnectionException("IO problems", e);
		}
	}

	/**
	 * Send a message to the server
	 * 
	 * @param msg Message in ASCII form
	 */
	@Override
	public void sendMessage(byte[] msg) throws ConnectionException
	{
		logger.log(Level.FINEST, "sending: {0}", msg);
		byte[] body = msg;

		// System.out.println(new String(msg));

		try {
			out.write(body);
			out.flush();
		} catch (IOException e) {
			shutDown = true;
			throw new ConnectionException(
					"Error writing to socket, shutting down...", e);
		}
	}

	/**
	 * Starts the endless loop of receiving messages and notifying observers
	 */
	@Override
	public void startReceiveLoop() throws ConnectionException
	{
		try {
			while (!shutDown) {
				byte[] currentMessage = receiveMessage();

				logger.log(Level.FINER, "Received message from server");
				logger.log(Level.FINEST, "message: {0}", currentMessage);

				// trigger main update cycle
				observer.onStateChange(currentMessage);

				// debug information
				endTime = System.nanoTime();
				logger.log(Level.FINE, "Message processed in (ns): "
						+ (endTime - startTime));
			}
		} finally {
			closeConnection();
			shutDown = false;
		}
	}

	/*
	 * Receive server message
	 */
	protected byte[] receiveMessage() throws ConnectionException
	{

		try {
			startTime = System.nanoTime();

			int length = 16384;
			byte[] buffer = new byte[length];
			int total = in.read(buffer, 0, length);

			if (total < 0) {
				// server was shutdown
				shutDown = true;
				throw new ConnectionException("Server shut down");
			}

			// System.out.println(new String(result));

			byte[] result = new byte[total];
			System.arraycopy(buffer, 0, result, 0, total);
			return result;

		} catch (IOException e) {
			shutDown = true;
			throw new ConnectionException(
					"Error when reading from socket, closing down...", e);
		}
	}

	/*
	 * Close server connection
	 */
	public void closeConnection()
	{
		try {
			connected = false;
			in.close();
			out.close();
			socket.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
